function initItems()
    Items = {}
    Items.current = {}
    Items.selection = nil
    Items.snaps = {}
    Items.snapBox = Box(0,0,50,50)
    Items.floor = screenSize.y * 0.6
    Items.gravity = 5000
    Items.minScale = 0.666

    initPics()
end

function resetItems()
    Items.current = {}
    Items.selection = nil
    for _,itemName in ipairs(Combi.initialItems) do
        addItem(itemName)
    end

    addAllSnaps()
end

function addAllSnaps()
    Items.snaps = {}
    Items.pentaSnaps = {}

    for _,snapPos in ipairs(ShelfPositions) do
        addSnap(snapPos)
    end

    for _,snapPos in ipairs(PentagramPositions) do
        addSnap(snapPos, "pentagram")
    end

    addSnap(CentralPosition, "central")

    repositionItemsInShelves()
end

function addItem(itemName)
    local item = {
        foot = Vector(math.random()*(screenSize.x - 160), math.random()*(screenSize.y - 160)),
        size = Vector(160,160),
        scale = Items.minScale,
        getBox = function(self)
            local sz = self.size * self.scale
            return Box(self.foot.x - sz.x * 0.5, self.foot.y - sz.y, self.foot.x + sz.x * 0.5, self.foot.y)
        end,
        acc = Vector(0,0),
        vel = Vector(0,0),
        color = {(math.random()*0.5+0.25)*255, (math.random()*0.5+0.25)*255, (math.random()*0.5+0.25)*255},
        name = itemName,
        opacity = 1,
        ofs = Vector(0,0),
        shineOpacity = 1,
        }
    table.insert(Items.current, item)
    return item
end

function addSnap(pos, isSpecial)
    local snap = {
        foot = pos,
        box = Items.snapBox + pos - Vector(Items.snapBox:width()*0.5, Items.snapBox:height()),
        isSpecial = isSpecial,
        }

    table.insert(Items.snaps, snap)
    if isSpecial == "pentagram" then
        table.insert(Items.pentaSnaps, snap)
    elseif isSpecial == "central" then
        Items.centralSnap = snap
    end
end

function repositionItemsInShelves()
    local availableShelves = {}
    for _,snap in ipairs(Items.snaps) do
        if not snap.isSpecial then table.insert(availableShelves, snap) end
    end
    for _,item in ipairs(Items.current) do
        local snap = table.remove(availableShelves, math.random(#availableShelves))
        putInSnap(item, snap)
    end
end

function updateItems(dt)
    local itemMoved = false
    for _,item in ipairs(Items.current) do
        local oldPosY = item.foot.y
        updateItem(item, dt)
        if not itemMoved and item.foot.y ~= oldPosY then
            itemMoved = true
        end
    end

    -- prevent flicking when z-sorting
    if itemMoved then
        local posDict = {}
        for i,item in ipairs(Items.current) do
            if posDict[item.foot.y] then
                item.foot.y = item.foot.y + i/#Items.current
            end
            posDict[item.foot.y] = true
        end
        sortItems()
    end


    -- mouse
    if not Rituals.active then
        local mpos = Vector(love.mouse.getX() / screenScale.x, love.mouse.getY() / screenScale.y)
        updateSelection(mpos, love.mouse.isDown('l'))
    end
end

function updateSelection(mpos,isDown)
    if not Items.selection and isDown then
        local item = getItemUnder(mpos)
        if item then
            takeItem(item, mpos)
        end
    elseif Items.selection and not isDown then
        releaseItem(Items.selection.target)
        Items.selection = nil
    elseif isDown then
        Items.selection.target.foot = mpos + Items.selection.opos
    end
end

function takeItem(item, mpos)
    Items.selection = {
        opos = item.foot - mpos,
        target = item
    }

    -- reset movement
    item.acc = Vector(0,0)
    item.vel = Vector(0,0)

    -- playSound(Sfx.pick)
end


function releaseItem(item)
    local function itemInSnap(snap)
        return snap.box:intersects(item:getBox())
    end
    local function getSnap()
        for _,snap in ipairs(Items.snaps) do
            if itemInSnap(snap) and (not snap.hasItem or snap.hasItem == item) then return snap end
        end
    end

    local snap = getSnap()
    if snap then
        -- snap into position
        putInSnap(item, snap)
        if snap.isSpecial == "pentagram" then
            playSound(Sfx.place)
        end

        if snap.isSpecial == "pentagram" and not Items.centralSnap.hasItem then
            local res = checkInvocation()
            if res then
                if not manageSpecialActivation(res) then
                    local newItem = findItem(res)
                    if not newItem then
                        newItem = addItem(res)
                        putInSnap(newItem, Items.centralSnap)
                        appearRitual(newItem)
                        launchRitual()
                    end
                end

                -- if not newItem.hasSnap or newItem.hasSnap.isSpecial ~= "pentagram" then

                -- end
            end
        end

    elseif item.hasSnap then
        -- fall down
        dropFromSnam(item)
    end
end

function manageSpecialActivation(itemName)
    if itemName == "Santa" then
        launchSantaRitual()
        return true
    elseif itemName == "Goat" then
        launchGoatRitual()
        return true
    elseif itemName == "Satan" then
        launchDevilRitual()
        return true
    end

    return false
end

function findInList(list, cond)
    for _,elem in ipairs(list) do
        if cond(elem) then return elem end
    end
    return nil
end

function findItem(itemName)
    return findInList(Items.current, function(item) return item.name == itemName end)
end

function dropFromSnam(item)
    item.hasSnap.hasItem = nil
    item.hasSnap = nil
end

function putInSnap(item, snap)
    item.foot = snap.foot:copy()

    if item.hasSnap then
        item.hasSnap.hasItem = nil
    end
    item.hasSnap = snap
    snap.hasItem = item
end

function isItemOnFloor(item)
    return item.foot.y >= Items.floor
end

function getItemUnder(mpos)
    for _,item in ipairs(Items.current) do
        local npos = (mpos - item:getBox():topLeft()) / item.scale
        if npos.x >= 0 and npos.y >= 0 and npos.x <= item.size.x and npos.y <= item.size.y then
            return item
        end
    end
    return nil
end

function drawItems()
    -- love.graphics.setColor(255, 0, 0)
    -- love.graphics.line(0, Items.floor, screenSize.x, Items.floor)

    -- love.graphics.setColor(0,255,0)
    -- for _,snap in ipairs(Items.snaps) do
    --     love.graphics.rectangle("line", snap.box.x0, snap.box.y0, snap.box:width(), snap.box:height())
    -- end

    for _,item in ipairs(Items.current) do
        drawItem(item)
    end

end

function sortItems()
    table.sort(Items.current, function(a,b) return a.foot.y < b.foot.y end)
end

function updateItem(item, dt)
    -- fall
    if not item.hasSnap and not isItemOnFloor(item) and (not Items.selection or Items.selection.target ~= item) then
        item.acc.y = Items.gravity
        item.vel.y = item.vel.y + item.acc.y * dt
        item.foot.y = item.foot.y + item.vel.y * dt
    else
        if item.vel.y > 0 then
            item.foot.y = Items.floor
            playSound(Sfx.fall)
        end
        item.vel.y = 0
        item.acc.y = 0
    end

    -- shine
    local shine = checkShine(item.name)
    if shine and item.shineOpacity < 1 then
        item.shineOpacity = increaseExponential(dt, item.shineOpacity, 0.95)
    elseif not shine and item.shineOpacity > 0 then
        item.shineOpacity = decreaseExponential(dt, item.shineOpacity, 0.95)
    end

    -- perspective
    if item.foot.y < Items.floor then
        item.scale = Items.minScale
    else
        item.scale = (1-Items.minScale) * (item.foot.y - Items.floor) / (screenSize.y - Items.floor) + Items.minScale
    end

    sortItems()
end

function increaseExponential(dt, var, amount)
    if var < 1 then
        var = 1 - (1 - var) * math.pow(amount, 60*dt)
        if var > 0.999 then
            var = 1
        end
    end
    return var
end

function decreaseExponential(dt, var, amount)
    if var > 0 then
        var = var * math.pow(amount, 60*dt)
        if var < 0.001 then
            var = 0
        end
    end
    return var
end


function drawItem(item)
    if Items.animMap[item.name] then
        return
    end

    -- local box = item:getBox() + item.ofs
    -- if item.shineOpacity > 0 then
    --     local shine = checkShine(item.name)
    --     if shine == 1 then
    --         love.graphics.setColor(89, 159, 255, item.opacity * item.shineOpacity * 255)
    --     else
    --         love.graphics.setColor(210, 0, 220, item.opacity * item.shineOpacity * 255)
    --     end
    --     love.graphics.rectangle("fill", box.x0-4, box.y0-4, box:width()+8, box:height()+8)
    -- end

    -- local r,g,b = unpack(item.color)
    -- love.graphics.setColor(r,g,b, item.opacity * 255)
    -- love.graphics.rectangle("fill", box.x0, box.y0, box:width(), box:height())

    -- love.graphics.setColor(0,0,0, item.opacity * 255)
    -- love.graphics.print(item.name, box.x0, box.y0)
end

function checkInvocation()
    local invoItems = {}
    for _,penta in ipairs(Items.pentaSnaps) do
        if not penta.hasItem then
            break
        end
        table.insert(invoItems, penta.hasItem.name)
    end

    if #Items.pentaSnaps == #invoItems then
        local res = getCombiResult(invoItems)
        if res then
            local isNew = consumeCombi(res)
            if isNew then
                -- markShinesFor(res[1])
                combineRitual(res[1])
                vibrateRitual(invoItems)
            end
            return res[2]
        else
            checkNearMiss(invoItems)
            return nil
        end
        -- print(res and res[2] or "none")
    else
        -- print("not enough")
    end

    return nil
end

function checkNearMiss(nameList)
    local candidates = getNearMiss(nameList)
    if candidates and #candidates > 2 then
        combineRitual({})
        vibrateRitual(candidates)
        appearRitual({})
        launchHintRitual()
    end
end

-- function markShinesFor(nameList)
--     for _,itemName in ipairs(nameList) do
--         markShine(itemName)
--     end
-- end
